@@ -1,5 +1,6 @@
 # -*- coding: utf-8 -*-
 
+import py
 import pytest
 
 import sys
@@ -7,7 +8,7 @@ import ssl
 import socket
 import threading
 import datetime
-from concurrent.futures import ThreadPoolExecutor
+from concurrent.futures import ThreadPoolExecutor  # type: ignore[import]
 
 from ipaddress import IPv4Address, IPv6Address, IPv4Network, IPv6Network
 
@@ -16,19 +17,27 @@ from cryptography.hazmat.backends import default_backend
 from cryptography.hazmat.primitives.serialization import (
     Encoding, PublicFormat, load_pem_private_key)
 
-import OpenSSL
-import service_identity.pyopenssl
+import OpenSSL.SSL
+import service_identity.pyopenssl  # type: ignore[import]
 
 import trustme
-from trustme import CA
+from trustme import CA, LeafCert
+
+TYPE_CHECKING = False
+if TYPE_CHECKING:  # pragma: no cover
+    from typing import Callable, Optional, Text, Union
+
+    SslSocket = Union[ssl.SSLSocket, OpenSSL.SSL.Connection]
 
 
 def _path_length(ca_cert):
+    # type: (x509.Certificate) -> Optional[int]
     bc = ca_cert.extensions.get_extension_for_class(x509.BasicConstraints)
     return bc.value.path_length
 
 
 def assert_is_ca(ca_cert):
+    # type: (x509.Certificate) -> None
     bc = ca_cert.extensions.get_extension_for_class(x509.BasicConstraints)
     assert bc.value.ca is True
     assert bc.critical is True
@@ -38,7 +47,24 @@ def assert_is_ca(ca_cert):
     assert ku.value.crl_sign is True
     assert ku.critical is True
 
-    eku = ca_cert.extensions.get_extension_for_class(x509.ExtendedKeyUsage)
+    with pytest.raises(x509.ExtensionNotFound):
+        ca_cert.extensions.get_extension_for_class(x509.ExtendedKeyUsage)
+
+
+def assert_is_leaf(leaf_cert):
+    # type: (x509.Certificate) -> None
+    bc = leaf_cert.extensions.get_extension_for_class(x509.BasicConstraints)
+    assert bc.value.ca is False
+    assert bc.critical is True
+
+    ku = leaf_cert.extensions.get_extension_for_class(x509.KeyUsage)
+    assert ku.value.digital_signature is True
+    assert ku.value.key_encipherment is True
+    assert ku.value.key_cert_sign is False
+    assert ku.value.crl_sign is False
+    assert ku.critical is True
+
+    eku = leaf_cert.extensions.get_extension_for_class(x509.ExtendedKeyUsage)
     assert eku.value == x509.ExtendedKeyUsage([
         x509.oid.ExtendedKeyUsageOID.CLIENT_AUTH,
         x509.oid.ExtendedKeyUsageOID.SERVER_AUTH,
@@ -48,6 +74,7 @@ def assert_is_ca(ca_cert):
 
 
 def test_basics():
+    # type: () -> None
     ca = CA()
 
     today = datetime.datetime.today()
@@ -88,6 +115,7 @@ def test_basics():
 
     assert server_cert.not_valid_before <= today <= server_cert.not_valid_after
     assert server_cert.issuer == ca_cert.subject
+    assert_is_leaf(server_cert)
 
     san = server_cert.extensions.get_extension_for_class(x509.SubjectAlternativeName)
     hostnames = san.value.get_values_for_type(x509.DNSName)
@@ -95,6 +123,7 @@ def test_basics():
 
 
 def test_ca_custom_names():
+    # type: () -> None
     ca = CA(
         organization_name=u'python-trio',
         organization_unit_name=u'trustme',
@@ -115,6 +144,7 @@ def test_ca_custom_names():
 
 
 def test_issue_cert_custom_names():
+    # type: () -> None
     ca = CA()
     leaf_cert = ca.issue_cert(
         u'example.org',
@@ -136,7 +166,30 @@ def test_issue_cert_custom_names():
     })
 
 
+def test_issue_cert_custom_not_after():
+    # type: () -> None
+     now = datetime.datetime.now()
+     expires = datetime.datetime(2025, 12, 1, 8, 10, 10)
+     ca = CA()
+
+     leaf_cert = ca.issue_cert(
+         u'example.org',
+         organization_name=u'python-trio',
+         organization_unit_name=u'trustme',
+         not_after=expires,
+     )
+
+     cert = x509.load_pem_x509_certificate(
+         leaf_cert.cert_chain_pems[0].bytes(),
+         default_backend(),
+     )
+
+     for t in ["year", "month", "day", "hour", "minute", "second"]:
+         assert getattr(cert.not_valid_after, t) == getattr(expires, t)
+
+
 def test_intermediate():
+    # type: () -> None
     ca = CA()
     ca_cert = x509.load_pem_x509_certificate(
         ca.cert_pem.bytes(), default_backend())
@@ -156,9 +209,11 @@ def test_intermediate():
     child_server_cert = x509.load_pem_x509_certificate(
         child_server.cert_chain_pems[0].bytes(), default_backend())
     assert child_server_cert.issuer == child_ca_cert.subject
+    assert_is_leaf(child_server_cert)
 
 
 def test_path_length():
+    # type: () -> None
     ca = CA()
     ca_cert = x509.load_pem_x509_certificate(
         ca.cert_pem.bytes(), default_backend())
@@ -177,17 +232,19 @@ def test_path_length():
 
 
 def test_unrecognized_context_type():
+    # type: () -> None
     ca = CA()
     server = ca.issue_cert(u"test-1.example.org")
 
     with pytest.raises(TypeError):
-        ca.configure_trust(None)
+        ca.configure_trust(None)  # type: ignore[arg-type]
 
     with pytest.raises(TypeError):
-        server.configure_cert(None)
+        server.configure_cert(None)  # type: ignore[arg-type]
 
 
 def test_blob(tmpdir):
+    # type: (py.path.local) -> None
     test_data = b"xyzzy"
     b = trustme.Blob(test_data)
 
@@ -223,6 +280,7 @@ def test_blob(tmpdir):
             assert f.read() == test_data
 
 def test_ca_from_pem(tmpdir):
+    # type: (py.path.local) -> None
     ca1 = trustme.CA()
     ca2 = trustme.CA.from_pem(ca1.cert_pem.bytes(), ca1.private_key_pem.bytes())
     assert ca1._certificate == ca2._certificate
@@ -230,13 +288,16 @@ def test_ca_from_pem(tmpdir):
 
 
 def check_connection_end_to_end(wrap_client, wrap_server):
+    # type: (Callable[[CA, socket.socket, Text], SslSocket], Callable[[LeafCert, socket.socket], SslSocket]) -> None
     # Client side
     def fake_ssl_client(ca, raw_client_sock, hostname):
+        # type: (CA, socket.socket, Text) -> None
         try:
             wrapped_client_sock = wrap_client(ca, raw_client_sock, hostname)
             # Send and receive some data to prove the connection is good
             wrapped_client_sock.send(b"x")
             assert wrapped_client_sock.recv(1) == b"y"
+            wrapped_client_sock.close()
         except:  # pragma: no cover
             sys.excepthook(*sys.exc_info())
             raise
@@ -245,11 +306,13 @@ def check_connection_end_to_end(wrap_client, wrap_server):
 
     # Server side
     def fake_ssl_server(server_cert, raw_server_sock):
+        # type: (LeafCert, socket.socket) -> None
         try:
             wrapped_server_sock = wrap_server(server_cert, raw_server_sock)
             # Prove that we're connected
             assert wrapped_server_sock.recv(1) == b"x"
             wrapped_server_sock.send(b"y")
+            wrapped_server_sock.close()
         except:  # pragma: no cover
             sys.excepthook(*sys.exc_info())
             raise
@@ -257,6 +320,7 @@ def check_connection_end_to_end(wrap_client, wrap_server):
             raw_server_sock.close()
 
     def doit(ca, hostname, server_cert):
+        # type: (CA, Text, LeafCert) -> None
         # socketpair and ssl don't work together on py2, because... reasons.
         # So we need to do this the hard way.
         listener = socket.socket()
@@ -297,18 +361,23 @@ def check_connection_end_to_end(wrap_client, wrap_server):
 
 
 def test_stdlib_end_to_end():
+    # type: () -> None
     def wrap_client(ca, raw_client_sock, hostname):
+        # type: (CA, socket.socket, Text) -> ssl.SSLSocket
         ctx = ssl.create_default_context()
         ca.configure_trust(ctx)
+        # Type ignore for Python 2: wants str, got unicode, but I guess unicode also works.
         wrapped_client_sock = ctx.wrap_socket(
-            raw_client_sock, server_hostname=hostname)
+            raw_client_sock, server_hostname=hostname)  # type: ignore[arg-type]
         print("Client got server cert:", wrapped_client_sock.getpeercert())
         peercert = wrapped_client_sock.getpeercert()
+        assert peercert is not None
         san = peercert["subjectAltName"]
         assert san == (("DNS", "my-test-host.example.org"),)
         return wrapped_client_sock
 
     def wrap_server(server_cert, raw_server_sock):
+        # type: (LeafCert, socket.socket) -> ssl.SSLSocket
         ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
         server_cert.configure_cert(ctx)
         wrapped_server_sock = ctx.wrap_socket(
@@ -320,12 +389,14 @@ def test_stdlib_end_to_end():
 
 
 def test_pyopenssl_end_to_end():
+    # type: () -> None
     def wrap_client(ca, raw_client_sock, hostname):
+        # type: (CA, socket.socket, Text) -> OpenSSL.SSL.Connection
         # Cribbed from example at
         #   https://service-identity.readthedocs.io/en/stable/api.html#service_identity.pyopenssl.verify_hostname
         ctx = OpenSSL.SSL.Context(OpenSSL.SSL.SSLv23_METHOD)
         ctx.set_verify(OpenSSL.SSL.VERIFY_PEER,
-                       lambda conn, cert, errno, depth, ok: ok)
+                       lambda conn, cert, errno, depth, ok: bool(ok))
         ca.configure_trust(ctx)
         conn = OpenSSL.SSL.Connection(ctx, raw_client_sock)
         conn.set_connect_state()
@@ -334,6 +405,7 @@ def test_pyopenssl_end_to_end():
         return conn
 
     def wrap_server(server_cert, raw_server_sock):
+        # type: (LeafCert, socket.socket) -> OpenSSL.SSL.Connection
         ctx = OpenSSL.SSL.Context(OpenSSL.SSL.SSLv23_METHOD)
         server_cert.configure_cert(ctx)
 
@@ -346,11 +418,12 @@ def test_pyopenssl_end_to_end():
 
 
 def test_identity_variants():
+    # type: () -> None
     ca = CA()
 
     for bad in [b"example.org", bytearray(b"example.org"), 123]:
         with pytest.raises(TypeError):
-            ca.issue_cert(bad)
+            ca.issue_cert(bad)  # type: ignore[arg-type]
 
     cases = {
         # Traditional ascii hostname
@@ -401,27 +474,30 @@ def test_identity_variants():
         san = cert.extensions.get_extension_for_class(
             x509.SubjectAlternativeName
         )
-        got = san.value[0]
+        assert_is_leaf(cert)
+        got = list(san.value)[0]
         assert got == expected
 
 
 def test_backcompat():
+    # type: () -> None
     ca = CA()
     # We can still use the old name
     ca.issue_server_cert(u"example.com")
 
 
 def test_CN():
+    # type: () -> None
     ca = CA()
 
     # Since we have to emulate kwonly args here, I guess we should test the
     # emulation logic
     with pytest.raises(TypeError):
-        ca.issue_cert(comon_nam=u"wrong kwarg name")
+        ca.issue_cert(comon_nam=u"wrong kwarg name")  # type: ignore[call-arg]
 
     # Must be unicode
     with pytest.raises(TypeError):
-        ca.issue_cert(common_name=b"bad kwarg value")
+        ca.issue_cert(common_name=b"bad kwarg value")  # type: ignore[arg-type]
 
     # Default is no common name
     pem = ca.issue_cert(u"example.com").cert_chain_pems[0].bytes()
@@ -445,7 +521,7 @@ def test_CN():
     san = cert.extensions.get_extension_for_class(
         x509.SubjectAlternativeName
     )
-    assert san.value[0] == x509.DNSName(u"example.com")
+    assert list(san.value)[0] == x509.DNSName(u"example.com")
     common_names = cert.subject.get_attributes_for_oid(
         x509.oid.NameOID.COMMON_NAME
     )
